local _, class = UnitClass("player")
if class ~= "HUNTER" then return end

local modName = "__marksmanship"
local mod = clcInfo:RegisterClassModule(modName)
local emod = clcInfo.env
local debug = clcInfo.debug
local db

-- status
local s1, s2
local s_gcd, s_ss, s_sf, s_mm, s_moc, s_rcss, s_boss, s_rf
local focus, maxf
local beast = 0

-- mode, 0 is Careful Aim, 1 is Standard, 2 is Kill Shot
local mode

local prio = {
	"se", -- serpent sting
	"moc", -- murder of crows
	"ssp", -- steady shot pair (steady focus)
	"cs", -- chimera shot
	"db", -- dire beast
	"ks", -- kill shot
	"rf", -- rapid fire
	"mm", -- aimed shot (master marksman)
	"ac", -- arcane shot
	"am", -- aimed shot
	"ss", -- steady shot
}

local function UseFocus(f)
	focus = max(0, focus - f)
end

local actions = {
	se = {
		id = emod.spells["Serpent Sting"],
		GetCD = function()
			if mode == 0 or s_ss > 0 then
				return 100
			end
			return 0
		end,
		UpdateStatus = function()
			emod:SetTime(1.0)
			UseFocus(25)
			s_ss = 15
		end,
		info = "Serpent Sting",
	},

	moc = {
		id = emod.spells["A Murder of Crows"],
		GetCD = function()
			-- don't show major cooldowns for trash
			if not s_boss then return 100 end
			if s1 == emod.spells["A Murder of Crows"] then return 100 end
			if s_moc > 0 then return 100 end
			return emod:GetCooldown(emod.spells["A Murder of Crows"])
		end,
		UpdateStatus = function()
			emod:SetTime(1.0)
			UseFocus(60)
			s_moc = 30.0
		end,
		info = "A Murder of Crows",
	},

	ssp = {
		id = emod.spells["Steady Shot"],
		GetCD = function()
			if s_sf < emod:GetCastTime(emod.spells["Steady Shot"]) and s_rcss < 2 then
				return 0
			end
			-- unsure if we should attempt to recast before
			-- buff drops
			return 100
		end,
		UpdateStatus = function()
			emod:SetTime(emod:GetCastTime(emod.spells["Steady Shot"]))
			focus = min(focus + 14, maxf)
			s_rcss = s_rcss + 1
		end,
		info = "Steady Shot (Steady Focus)",
	},

	cs = {
		id = emod.spells["Chimera Shot"],
		GetCD = function()
			if mode == 0 then return 100 end
			if s1 == emod.spells["Chimera Shot"] then return 100 end
			return emod:GetCooldown(emod.spells["Chimera Shot"])
		end,
		UpdateStatus = function()
			emod:SetTime(1.0)
			UseFocus(45)
		end,
		info = "Chimera Shot",
	},

	db = {
		id = emod.spells["Dire Beast"],
		GetCD = function()
			if s1 == emod.spells["Dire Beast"] then return 100 end
			return emod:GetCooldown(emod.spells["Dire Beast"])
		end,
		UpdateStatus = function()
			emod:SetTime(1.0)
		end,
		info = "Dire Beast",
	},

	ks = {
		id = emod.spells["Kill Shot"],
		GetCD = function()
			if s1 == emod.spells["Kill Shot"] then return 100 end
			if mode == 2 then return emod:GetCooldown(emod.spells["Kill Shot"]) end
			return 100
		end,
		UpdateStatus = function()
			emod:SetTime(1.0)
		end,
		info = "Kill Shot",
	},

	rf = {
		id = emod.spells["Rapid Fire"],
		GetCD = function()
			-- don't show major cooldowns for trash
			if s1 == emod.spells["Rapid Fire"] then return 100 end
			if not s_boss then return 100 end
			if s_rf > 0 then return 100 end
			return emod:GetCooldown(emod.spells["Rapid Fire"])
		end,
		UpdateStatus = function()
			s_rf = 15.0
		end,
		info = "Rapid Fire",
	},

	rdy = {
		id = emod.spells["Readiness"],
		GetCD = function()
			-- don't show major cooldowns for trash
			if not s_boss then return 100 end
			if s1 == emod.spells["Readiness"] then return 100 end
			return emod:GetCooldown(emod.spells["Readiness"])
		end,
		UpdateStatus = function()
			emod:UseReadiness()
		end,
		info = "Readiness",
	},

	mm = {
		id = emod.spells["Aimed Shot"],
		GetCD = function()
			if s_mm > 0 then
				return 0
			end
			return 100
		end,
		UpdateStatus = function()
			emod:SetTime(1.0)
			s_mm = 0
		end,
		info = "Aimed Shot (MM)",
	},

	ac = {
		id = emod.spells["Arcane Shot"],
		GetCD = function()
			-- aimed shot needs to replace arcane if it's cast time
			-- is under 1.4s
			if mode == 0 then return 100 end
			if emod:GetCastTime(emod.spells["Aimed Shot"]) <= 1.4 then return 100 end
			-- if we can't fit another steady in before chimera, we shouldn't spend
			if focus - 25 < 45 and emod:GetCastTime(emod.spells["Steady Shot"]) > emod:GetCooldown(emod.spells["Chimera Shot"]) then
				return 100
			end
			if focus >= 25 then return 0 end
			return 100
		end,
		UpdateStatus = function()
			emod:SetTime(1.0)
			UseFocus(25)
		end,
		info = "Arcane Shot",
	},

	am = {
		id = emod.spells["Aimed Shot"],
		GetCD = function()
			-- use aimed if it's cast time is <1.4s
			if mode > 0 and emod:GetCastTime(emod.spells["Aimed Shot"]) > 1.4 then return 100 end
			if focus >= 50 then return 0 end
			return 100
		end,
		UpdateStatus = function()
			emod:SetTime(emod:GetCastTime(emod.spells["Aimed Shot"]))
			UseFocus(50)
		end,
		info = "Aimed Shot",
	},

	ss = {
		id = emod.spells["Steady Shot"],
		GetCD = function() return emod:GetCastTime(emod.spells["Steady Shot"]) end,
		UpdateStatus = function()
			emod:SetTime(emod:GetCastTime(emod.spells["Steady Shot"]))
			focus = min(focus + 14, maxf)
			s_rcss = s_rcss + 1
		end,
		info = "Steady Shot",
	},
}

local function GetBuff(buff)
	local left = 0
	local __, __, __, __, __, __, expires = UnitBuff("player", buff, nil, "PLAYER")
	if expires then
		left = max(0, expires - s_ctime - s_gcd)
	end
	return left
end

local function GetTargetDebuff(debuff)
	local left = 0
	local name, __, __, __, __, __, expires, __, __, __, __ = UnitAura("target", debuff, nil, "PLAYER|HARMFUL")
	if expires and expires > 0 then
		left = max(0, expires - s_ctime, s_gcd)
	elseif name then
		-- debuffs without times (murder??!)
		left = 100
	end
	return left
end

local function GetStatus()
	s_ctime = emod:GetTime()

	-- gcd
	s_gcd = emod:GetGlobalCooldown()

	-- update mode
	local hp = UnitHealth("target")
	local hpmax = UnitHealthMax("target")
	if hp > 0 and hpmax > 0 then
		local pct = (hp / hpmax) * 100.0
		if pct >= 90.0 then mode = 0 end
		if pct < 90.0 and pct > 20.0 then mode = 1 end
		if pct <= 20.0 then mode = 2 end
	else
		mode = 0
	end

	-- -- update info
	s_ss = GetTargetDebuff(emod.auras["Serpent Sting"])
	s_moc = GetTargetDebuff(emod.auras["A Murder of Crows"])
	s_sf = GetBuff(emod.auras["Steady Focus"])
	s_mm = GetBuff(emod.auras["Fire!"])
	s_rf = GetBuff(emod.auras["Rapid Fire"])
	focus = UnitPower("player")
	maxf = UnitPowerMax("player")

	-- take a guess at whether we're facing a boss
	if UnitClassification("target") == "worldboss" then
		s_boss = true
	elseif UnitClassification("target") == "elite" and UnitHealth("target") > 4000000 then
		s_boss = true
	else
		s_boss = false
	end

	-- capture the number of recent steady shots for stability of prediction
	s_rcss = emod:GetRecentSteadyShotCount()
end

local function GetNextAction()
	local q = prio
	local n = #q

	for i=1, n do
		local action = actions[q[i]]
		local cd = action.GetCD()
		if cd == 0 then
			return action.id, q[i]
		end
		action.cd = cd
	end

	local minQ = 1
	local minCD = actions[q[1]].cd
	for i=2, n do
		local action = actions[q[i]]
		if minCD > action.cd then
			minCD = action.cd
			minQ = i
		end
	end
	return actions[q[minQ]].id, q[minQ]
end

local function MarksRotation()
	s1 = nil
	GetStatus()
	if debug.enabled then
		debug:Clear()
		debug:AddBoth("mode", mode)
		debug:AddBoth("s_ss", s_ss)
		debug:AddBoth("s_mm", s_mm)
		debug:AddBoth("s_sf", s_sf)
		debug:AddBoth("focus", focus)
		debug:AddBoth("s_moc", s_moc)
		debug:AddBoth("recent_ss", emod:GetRecentSteadyShotCount())
		debug:AddBoth("beast", GetTime() - beast)
	end

	local action
	s1, action = GetNextAction()
	if debug.enabled then
		debug:AddBoth("s1", action)
	end

	s_otime = s_ctime
	-- each action needs to update s_ctime for the time taken to cast
	actions[action].UpdateStatus()

	s_otime = s_ctime - s_otime
	s_ss = max(0, s_ss - s_otime)
	s_mm = max(0, s_mm - s_otime)

	-- now calculate how much generic regen we have made
	local __, active = GetPowerRegen()
	focus = min(maxf, focus + (active * s_otime))

	-- if dire beast is active we have to guess at it's possible regen
	if beast and beast > 0 and (GetTime() - beast) < 15.0 then
		local s_haste = 1 + GetRangedHaste("player") / 100
		-- DB has base 2s attack speed, adjusted by haste
		local regen = (15 / (2.0 * s_haste * 5))
		focus = min(maxf, focus + (regen * s_otime))
	end

	if debug.enabled then
		debug:AddBoth("s_ss", s_ss)
		debug:AddBoth("s_mm", s_mm)
		debug:AddBoth("focus", focus)
	end

	s2, action = GetNextAction()
	if debug.enabled then
		debug:AddBoth("s2", action)
	end
end

function emod.IconMarks1(...)
	MarksRotation(...)
	if secondarySkill then secondarySkill:DoUpdate() end
	return emod.IconSpell(s1, true)
end

local function SecondaryExec()
	return emod.IconSpell(s2, true)
end

local function ExecCleanup2()
	secondarySkill = nil
end

function emod.IconMarks2(...)
	secondarySkill = emod.___e
	secondarySkill:SetScript("OnUpdate", nil)
	secondarySkill.exec = SecondaryExec
	secondarySkill.ExecCleanup = ExecCleanup2
end
